/*

      DirectSound3D
==================================================================

   DirectSound,    Priority,
     .   
      .

       
 ,     
.
             
*/





#pragma warning (disable: 4115 4201 4214 4514 4127 4702)


#if _DEBUG

  #pragma warning (disable: 4505)

#endif


#define STRICT                  //   

#define INITGUID                //    GUID

#include <windows.h>            //   Windows
#include <commctrl.h>           //    

#include "dsound.h"             // DirectSound 7.0

#include <math.h>               //  
#include <malloc.h>             //   
#include <stdio.h>              //  /

#include "resource.h"           //  





//
// #########################################################################
//
// 
// 
// #########################################################################
//





#define UNREF (void)





//
// #########################################################################
//
// 
// 
// #########################################################################
//





const int SampleRate = 22050;   //   


const D3DVALUE MinX = -02.0f;   //  X
const D3DVALUE MaxX =  02.0f;   //  X

const D3DVALUE MinY = -02.0f;   //  Y
const D3DVALUE MaxY =  02.0f;   //  Y

const D3DVALUE MinZ = -02.0f;   //  Z
const D3DVALUE MaxZ =  02.0f;   //  Z


const D3DVALUE SliderScale = 1000.0f; //  




//
// #########################################################################
//
// 
//
// #########################################################################
//


IDirectSound *Out;              //   

IDirectSoundBuffer *Primary;    //  
IDirectSound3DListener *Listener; // 


IDirectSoundBuffer *SndBuf;     //    
IDirectSound3DBuffer *SndSrc;   //   

HWND DlgWin;                    //   

BOOL Active;                    //  

UINT NDevs;                     //    

GUID *DevGuids [32];            //  

D3DVALUE X, Y, Z;               //  





//
// #########################################################################
//
// 
//
// #########################################################################
//





//
//    
// ==========================
//

static void Error (LPCTSTR Format, ...) {

  char Str [256];

  va_list Args;

  va_start (Args, Format);
  vsprintf (Str, Format, Args);
  va_end (Args);

  MessageBox (DlgWin, Str, "", MB_ICONERROR | MB_OK);

}





//
//      
// ======================================================
//

static LONG DlgMsg (int Id, UINT Msg, LPARAM lp, WPARAM wp) {

  return SendDlgItemMessage (DlgWin, Id, Msg, lp, wp);

}




//
//    ComboBox 
// ====================================
//

static void AddStringToCB (UINT Ctl, LPCTSTR Str) {

  DlgMsg (Ctl,  CB_ADDSTRING,  0, (LPARAM)Str);

}





//
// /  
// ==================================
//

static void EnableCtl (UINT Ctl, BOOL Enable = TRUE) {

  EnableWindow (GetDlgItem (DlgWin, Ctl), Enable);

}




//
//   
// ==========================
//

static void SetSliderRange (UINT Id, D3DVALUE Min, D3DVALUE Max) {

  DlgMsg (
    Id,
    TBM_SETRANGE,
    TRUE,
    MAKELONG (Min * SliderScale, Max * SliderScale)
  );

  DlgMsg (
    Id,
    TBM_SETPOS,
    TRUE,
    (DlgMsg (Id, TBM_GETRANGEMAX, 0, 0) + DlgMsg (Id, TBM_GETRANGEMIN, 0, 0)) / 2
  );

}





//
//    
// =====================================
//

static D3DVALUE GetSliderPos (UINT Id) {

  LONG Pos = DlgMsg (Id, TBM_GETPOS, 0, 0);

  if (GetWindowLong (GetDlgItem (DlgWin, Id), GWL_STYLE) & TBS_VERT) {

    Pos = -Pos;                 //     

  }

  return D3DVALUE (Pos) / SliderScale;

}





//
//   
// =======================
//

static void UpdateCoords (void) {

  char Str [30];

  sprintf (Str, "X=% 06.2f", X);
  SetWindowText (GetDlgItem (DlgWin, IDC_ValueX), Str);

  sprintf (Str, "Y=% 06.2f", Y);
  SetWindowText (GetDlgItem (DlgWin, IDC_ValueY), Str);

  sprintf (Str, "Z=% 06.2f", Z);
  SetWindowText (GetDlgItem (DlgWin, IDC_ValueZ), Str);

  if (SndSrc) {

    SndSrc->SetPosition (X, Y, Z, DS3D_IMMEDIATE);

  }

}





//
//  
// ====================
//

static void CloseDevices (void) {

  //
  //  
  //

  if (SndBuf) {

    SndBuf->Stop ();

  }


  //
  //   
  //

  if (SndSrc) {

    SndSrc->Release ();
    SndSrc = NULL;

  }

  if (SndBuf) {

    SndBuf->Release ();
    SndBuf = NULL;

  }

  if (Listener) {

    Listener->Release ();
    Listener = NULL;

  }

  if (Out) {

    Out->Release ();
    Out = NULL;

  }

}





//
//  
// ====================
//

static BOOL OpenDevices (void) {

  HRESULT Res;

  do {

    //
    //    DirectSound
    //

    GUID *DevGuid = DevGuids [DlgMsg (IDC_OutDev, CB_GETCURSEL, 0, 0)];

    Res = DirectSoundCreate (DevGuid, &Out, NULL);

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    //
    //    
    //

    Res = Out->SetCooperativeLevel (DlgWin, DSSCL_PRIORITY);

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    //
    //   
    //

    BYTE SpkCfgs [] = {         //  

      DSSPEAKER_HEADPHONE,
      DSSPEAKER_MONO,
      DSSPEAKER_QUAD,
      DSSPEAKER_STEREO,
      DSSPEAKER_SURROUND,
      DSSPEAKER_5POINT1,

    };

    UINT Cfg = SpkCfgs [DlgMsg (IDC_SpkCfg, CB_GETCURSEL, 0, 0)];

    UINT Geom = 0;

    if (Cfg == DSSPEAKER_STEREO) {

      char Str [20];

      GetWindowText (GetDlgItem (DlgWin, IDC_SpkGeom), Str, sizeof (Str));

      char *End;

      Geom = strtoul (Str, &End, 10);

      if (*End && !isspace (*End) || Geom > 180) {

        Error ("  ");

        break;

      }

    }

    Res = Out->SetSpeakerConfig (DSSPEAKER_COMBINED (Cfg, Geom));

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    //
    //   
    //

    DSBUFFERDESC BD;

    memset (&BD, 0, sizeof (BD));

    BD.dwSize = sizeof (BD);
    BD.dwFlags = DSBCAPS_PRIMARYBUFFER | DSBCAPS_CTRL3D;

    Res = Out->CreateSoundBuffer (&BD, &Primary, NULL);

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    //
    //   
    //

    WAVEFORMATEX Fmt = {
      WAVE_FORMAT_PCM,        //  
      1,                      // 
      SampleRate,             //  
      SampleRate * 2,         //   
      2,                      //   
      16,                     //  
      0,                      //  
    };


    //
    //    
    //

    Res = Primary->SetFormat (&Fmt);

    if (FAILED (Res)) {

      Error ("     (0x%X)", Res);

      break;

    }


    //
    //   
    //

    Res = Primary->QueryInterface (
      IID_IDirectSound3DListener,
      (void **)&Listener
    );

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    //
    //   
    //

    HRSRC WaveData;

    WaveData = FindResource (
      NULL,
      (LPCTSTR)IDR_WaveData,
      (LPCTSTR)RT_RCDATA
    );


    //
    //    
    //

    memset (&BD, 0, sizeof (BD));

    BD.dwSize = sizeof (BD);

    BD.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRL3D | DSBCAPS_CTRLFREQUENCY;
    BD.dwBufferBytes = SizeofResource (NULL, WaveData);
    BD.lpwfxFormat = &Fmt;

    LPCGUID VirtAlgs [] = {
      &DS3DALG_NO_VIRTUALIZATION,
      &DS3DALG_HRTF_LIGHT,
      &DS3DALG_HRTF_FULL,
    };

    memcpy (
      &BD.guid3DAlgorithm,
      VirtAlgs [DlgMsg (IDC_Virt, CB_GETCURSEL, 0, 0)],
      sizeof (BD.guid3DAlgorithm)
    );

    Res = Out->CreateSoundBuffer ((DSBUFFERDESC *)&BD, &SndBuf, NULL);

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    //
    //     
    //

    BYTE *P1, *P2;
    DWORD S1, S2;

    Res = SndBuf->Lock (
      0,
      0,
      (void **)&P1,
      &S1,
      (void **)&P2,
      &S2,
      DSBLOCK_FROMWRITECURSOR | DSBLOCK_ENTIREBUFFER
    );

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }

    BYTE *WDP = (BYTE *)LoadResource (NULL, WaveData);

    memcpy (P1, WDP, S1);
    memcpy (P2, WDP + S1, S2);

    Res = SndBuf->Unlock (P1, S1, P2, S2);

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    //
    //    
    //

    Res = SndBuf->QueryInterface (
      IID_IDirectSound3DBuffer,
      (void **)&SndSrc
    );

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    //
    //  
    //

    SndSrc->SetMinDistance (0.01f, DS3D_DEFERRED);

    UpdateCoords ();


    //
    //  
    //

    Res = SndBuf->Play (0, 0, DSBPLAY_LOOPING);

    if (FAILED (Res)) {

      Error ("    (0x%X)", Res);

      break;

    }


    return TRUE;

  } while (0);

  CloseDevices ();

  return FALSE;
    
}





//
//    
// =================================
//

static BOOL CALLBACK DSEnum (
  GUID *Guid,
  LPCSTR Name,
  LPCSTR Drv,
  void *Context
  
) {

  DevGuids [NDevs++] = Guid;

  AddStringToCB (IDC_OutDev, Name);

  return TRUE;

  UNREF (Context, Drv);

}





//
//    
// =====================================
//

static BOOL CALLBACK DialogProc (HWND Win, UINT Msg, WPARAM wp, LPARAM lp) {

  HRESULT Res;

  switch (Msg) {





    //
    //   
    // ==============================
    //

    case WM_INITDIALOG: {

      DlgWin = Win;


      //
      //   
      //

      Res = DirectSoundEnumerate (DSEnum, NULL);

      if (FAILED (Res)) {

        Error ("    (0x%X)", Res);

        EndDialog (Win, 1);

        break;

      }

      DlgMsg (IDC_OutDev, CB_SETCURSEL, 0, 0);


      //
      //    
      //

      AddStringToCB (IDC_Virt, "");
      AddStringToCB (IDC_Virt, "");
      AddStringToCB (IDC_Virt, "");

      DlgMsg (IDC_Virt, CB_SETCURSEL, 2, 0);


      //
      //    
      //

      AddStringToCB (IDC_SpkCfg, "");
      AddStringToCB (IDC_SpkCfg, "");
      AddStringToCB (IDC_SpkCfg, "");
      AddStringToCB (IDC_SpkCfg, "");
      AddStringToCB (IDC_SpkCfg, "");
      AddStringToCB (IDC_SpkCfg, " Surround");
      AddStringToCB (IDC_SpkCfg, " 5point1");

      DlgMsg (IDC_SpkCfg, CB_SETCURSEL, 0, 0);


      //
      //     
      //

      AddStringToCB (IDC_SpkGeom, "5");
      AddStringToCB (IDC_SpkGeom, "10");
      AddStringToCB (IDC_SpkGeom, "20");
      AddStringToCB (IDC_SpkGeom, "180");

      DlgMsg (IDC_SpkGeom, CB_SETCURSEL, 1, 0);


      //
      //   
      //

      SetSliderRange (IDC_SliderX, MinX, MaxX);
      SetSliderRange (IDC_SliderY, MinY, MaxY);
      SetSliderRange (IDC_SliderZ, MinZ, MaxZ);

      UpdateCoords ();


      return TRUE;

    }





    //
    //   Windows
    // ===========================
    //

    case WM_COMMAND: {

      UINT CtrlId = LOWORD (wp);
      //UINT CtrlMsg = HIWORD (wp);
      
      switch (CtrlId) {
        

        //
        //     -     
        // =================================================================
        //

        case IDCANCEL: {

          if (Active) {

            break;

          }

          EndDialog (DlgWin, 0);

          return TRUE;

        }
        



        //
        //   /
        // ===============================
        //

        case IDC_StartStop: {

          EnableCtl (IDC_StartStop, FALSE); //  

          if (!Active) {

            EnableCtl (IDC_OutDev, FALSE);
            EnableCtl (IDC_Virt, FALSE);
            EnableCtl (IDC_SpkCfg, FALSE);
            EnableCtl (IDC_SpkGeom, FALSE);

            UpdateWindow (Win);  //    

            SetWindowText (GetDlgItem (DlgWin, IDC_StartStop), "");

            if (OpenDevices ()) {

              Active = TRUE;

            }

          } else {

            CloseDevices ();

            Active = FALSE;

          }

          if (!Active) {

            EnableCtl (IDC_OutDev);
            EnableCtl (IDC_Virt);
            EnableCtl (IDC_SpkCfg);
            EnableCtl (IDC_SpkGeom);

          }

          SetWindowText (
            GetDlgItem (DlgWin, IDC_StartStop),
            Active? "" : ""
          );

          EnableCtl (IDC_StartStop); //  

          break;
        
        }






      } // switch (CtrlId)

      break;





    } // WM_COMMAND





    //
    //    
    // ===============================
    //

    case WM_HSCROLL:
    case WM_VSCROLL: {

      X = GetSliderPos (IDC_SliderX);

      Y = GetSliderPos (IDC_SliderY);

      Z = GetSliderPos (IDC_SliderZ);

      UpdateCoords ();

      break;

    }





  } // switch (Msg)






  return FALSE;

  UNREF (lp);

}
      

//
//   Windows
// ========================
//

int WINAPI WinMain (
  HINSTANCE Instance,
  HINSTANCE PrevInstance,
  LPSTR Cmd,
  int CmdShow
) {

  int Result;
  

  //
  //     
  //

  InitCommonControls ();


  //
  //   
  //

  Result = DialogBox (
    Instance, 
    MAKEINTRESOURCE (IDD_Dialog),
    HWND_DESKTOP,
    DialogProc
  );


  return Result;

  UNREF (Instance, PrevInstance, Cmd, CmdShow);

}
